
---
title: "Dep Command-based Build Automation Tool"
geometry:
- margin=5mm
- bottom=12mm
- right=8mm
...

dep is a software build automation tool, similar in function to make, makepp, mk, tup and ninja.

It emphasises:

* A powerful, expressive, easy-to-use input language
* Robust, highly capable model and internal data structures
* Complete integration and handling of subprojects

## Overview

dep's language requires little more than the build commands that it must run. For example, this:

gcc -o fnord.o -c fnord.c

is a complete build rule in dep. With a careful reading of the command, dep can tell that it:

* writes fnord.o
* reads fnord.c
* reads all of the files that fnord.c includes, along with all of the files that they include, and
   so on recursively.

The little extra that dep needs is the syntax of the command-line options, which is provided by a
line such as:

syntax gcc -o:%out -c%findincludes -.\* %in

dep provides, but does not require, a syntax for include files. It looks like this:

fileinfo fnord.cpp OK 2022-11-22 12:34:56.789 123456 includes foo.h

(There is a little more information in there, namely a build state, a modification time, and a file
size).  But that is for a cached copy of the information.  Most often, dep simply uses the native
include syntax, by reading the source file directly and looking for lines such as:

  #include "foo.h"

dep can also easily handle multiple build projects, similar to make's recursive functionality, but
without requiring recursion. dep simply includes all of a subprojects rules, and runs them as
necessary. Usually they will be in different directories; dep will change the filenames as necessary,
and change directory before running any commands found there.

dep also runs its build commands in parallel whenever possible.

## A BRIEF DIS(CUSSION) OF OTHER TOOLS

So many build automation tools have become available over the years that there are now a lot of
alternatives to choose from: make, nmake, GNU make, mk, makepp, ninja, tup, ant, ivy, maven, gradle,
bazel, pants, buck, please and rake.

A lot of ideas have gone into them. These ideas have varied considerably in brilliance. In no
particular order, and without naming names, I'll go over some of the lowlights:

  * Some of these tools can automatially download and install library dependencies. However, there is
    considerable discussion on the Internet about how useful this feature actually is, and an awful
    lot of advice about on how to turn this particular feature off.
  * Some of these tools prescribe directory hierarchies that would make a taxonomist gasp. E.g.
    app/src/main/java comes up a lot, which directory (unsurprisingly) houses the vast bulk of
    the files shipped with a Java source application, including (somewhat suprisingly) many files that
    aren't exactly mainline.
  * Parallel execution is provided _sometimes_, and if so, as an _experimental_ feature sometimes;
    whereas building multiple-source-file software is, quite frankly, one of the most embarrassingly
    parallelizable problems OF ALL TIME.
  * For all the talk about performance, some of these tools are slow! How on earth do you compress a
    compile/link/install/run workflow into sixty separate steps? Even starting a daemon to speed
    things up takes ages, and once it is running, it can be faster to just build the whole thing with
    a script, from scratch, than to run the "build automation" tool.
  * The language-oriented tools that pay lip service to general-purposeness are, in practice, not in
    the least bit general-purpose.

## DEP vs GNU MAKE and some others

Actually, I will name one tool, GNU make, which is without doubt my favorite of the existing build
automation tools, and some others which are very closely related: original UNIX make, Plan 9 mk, and
makepp.

UNIX make is absolutely the grandfather of all these tools. Originating in 1976 (so 46 years old as of
2022) there is no doubt it is long in the tooth, and struggles to handle modern applications.

GNU make was originally released about 1988, largely as a workalike of UNIX make, but has since
made enormous improvements in functionality. It is my favorite of the existing tools, it is my
baseline for developing dep, and it is still improving: a recent addition is the -Oline option, which
eliminates mixing of output from parallel child processes.

Plan 9's mk perfected the algorithm for implicit rules. dep, on the other hand, abandons implict rules
entirely.  In dep, build rule construction is procedural and explicit.

(I am quite sure that makepp goes even further in extending UNIX make functionality, and which dep
partly parallels, but I have too limited personal experience with makepp to go into any detail.)

dep is deliberately a break from the UNIX make / GNU make / mk / makepp line. It is intended to be a
successor to the make family, but it is not compatible. There are two fundamental reasons for this:

  * The make tools require (a) as input sources, targets, and commands, which (b) may be implicitly
    constructed and (c) may be tweaked. Whereas dep requires (a) just the commands, which (b) may be
    procedurally constructed and (c) are automatically analysed to determine the source and target
    dependency relationships.
  * The make tools' fundamental data type is the string, whereas dep's fundamental datatype is the
    filename. This one change allows dep to handle multiple-subproject, multi-directory environments
    with much more grace and dexterity.

In addition, in comparison to GNU make, there are several things that dep can do that make can't.
These are:

  * Determine sources and targets by analysis of build commands
  * Determine include files by simple reading of the source files
  * Directly model include files as include files, instead of as additional source files
  * As necessary, compare file contents instead of just assuming that a rebuilt file has changed
  * Provide straightforward constructs such as loops and conditions in the source language

There are also things that dep does considerably better:

  * Subprojects (i.e. include foo/DepFile) are drawn directly into the project tree instead of using
    recursive make
  * Determination of include files directly from source files (i.e. #include "fnord.h")
  * Running commands without invoking a shell
  * Handling of filenames that contain whitespace

And there are things that dep simply makes easier:

  * Commands with multiple targets
  * Parallel execution
  * Multiple platforms

There are also some features of GNU make that dep deliberately does not attempt to replicate:

  * .INTERMEDIATE, .SECONDARY, .PRECIOUS
  * Suffix rules and implicit rules
  * :: syntax
  * Multiple build rules for a target
  * Reprocessing of variable assignments
  * Recursive make
  * Jobservers
  * VPATH

There is however one important feature that is sometimes omitted by other tools, that dep
wholeheartedly retains:

  * .PHONY (however dep knows this as %pseudo)

dep is written in C++, which is my preferred programming language because it is fast and expressive,
even if it is rather complex.

## Contact

Have you found a bug? Got a suggestion? Maybe it's working perfectly! Let me know how you go.

Email me (Matthew Rickard) at dep@depbuild.org

## Release Notes

### dep 0.3

dep 0.3 was released on 9 March 2023. Changes are:

  * Changed findIncludes mechanism to look at filenames instead of options such as -c, to handle
    instances such as gcc a.c b.o
  * Fixed bugs in quote mark parsing
  * Combined operations on rule and tree includes

### dep 0.2

dep 0.2 was released on 16 February 2023. Changes are:

  * Added sensing and invalidation of some misbuilt files (generally 0 byte
    files, but 8 bytes for .a files)
  * Added %outz for files that may validly be zero length
  * Fixed bugs in handling of rules with no inputs e.g. mkdir commands
  * Added existence dependencies
  * Changed -p option to -j
  * No need to put " > %out " in syntax lines any more, it is now always assumed
  * Fixed bugs in interpretation of nested for loops
  * Allowed braces on subsequent lines of for loops
  * Changed "if ... endif" to "if { }"
  * Fixed some errors in the generation of .DepFileCache and suppressed some others in readback
  * Improved line and column number reporting in DepFile syntax error messages
  * Added /etc/depsyntax file, which includes syntaxes of common commands
  * Fixed bug in export statements within rules: they are now run during execution, not during parsing

### dep 0.1

dep 0.1 was released on 7 December 2022.

## Copyright and Licence

Copyright (C) 2022 Matthew Rickard

dep is free software licenced under the GNU General Public Licence version 3.

